# metrics-importer

This service is responsible for importing metrics to coronerd's metrics tables
for i.e. stability score.  The flow is as follows:

- User creates a metric and metric_group on the coronerd side, via BPG.
- Frontend hits `/api/metrics-importer/projects/name/sources` to create a
  source, which represents a connection to an entity we can import from, i.e.
  Graphite.
- Frontend hits `/api/metrics-importer/projects/name/importers` to create
  an importer which contains a query, a link to a source, and some other scraping
  parameters.
- The service periodically scrapes the source and ships to coronerd.

If an importer errors too many times it gets disabled; they can be re-enabled
via PUT to `/.../projects/name/importers/id` with `{ "enabled": true }`.

A logging endpoint `/.../projects/name/logs?sourceId=bla&importerId=bla`
is provided; either one of `sourceId` or `importerId` must be specified.
Pagination is via
`limit` and `offset`.  We don't currently expose a total count of log messages.
Logs are cleared with configurable retentions; the default is 30 days for
everything that's not debug, and 1 hour for debug.

The API endpoints are registerd in `src/endpoints/index.ts`. To find out what
they accept, see the top of the source file defining the endpoint for the one
you're interested in, and it'll have `export class BlaRequest`. We use
class-validator, which provides relatively easy to read validation, and comments
are provided for anything special.

Responses to create requests are always `{ "id": "uuid" }` but may be extended
in future to include the entire entity. Get requests are the entity itself from
`/src/entities` converted to JSON, with time fields transformed to Unix time.
list requests are the same.

## Documentation of URL and query configuration

### Graphite

Graphite has a
[render URL API](https://graphite.readthedocs.io/en/latest/render_api.html)
which can be used to get data out in a number of formats. To configure a
Graphite scraper, the user should build a render URL. The source's URL is
everything but the query params, and the query for the importer the query
params themselves. Note that the source's URL needs to include `/render`. This
is because people tend to do unusual things with Graphite deployment. For
example, Grafana can act as a proxy at a subdomain of the Grafana
installation.

Graphite itself is unauthenticated by default, and all authentication methods
which we support for HTTP might be in use, either via custom Django
authentication settings or via Nginx proxies.

### Influx

Influx is a high-volume timeseries database.  Queries look like SQL:

```
select time, field as value from database...measurement where
time >= $time_start and time < $time_end
```

`$time_start` and `$time_end` are the start and end of the range as nanoseconds
since the Unix epoch and must be included in every query in order to limit the
time range. Metrics-importer will error when executing the scrape if they
aren't present as we might otherwise take in millions of datapoints or more.

Every query must produce 2 columns, `time`, and `value`.  Like with sql, it's
possible to alias with `expr as name`, so this works fine even with
aggregations.  Users should also downsample data when possible. A more
practical/realistic query example, using data generated by the test scripts:

```
select time, sum(test_field) as value from test..data
where time >= $time_start and time < $time_end
group by time(5m)
```

It is important to note that metrics-importer requests don't specify a
database. Influx allows explicitly specifying it as
`database.[retention].measurement`, which we rely on to avoid additional
configuration on our end. So in the above, `test..data` means test database,
measurement `data`, and use the default retention policy. If the retention
policy isn't specified, `..` is mandatory and not a typo.

Additional where conditions are fine and expected usage, but depend on
customer data.

### Postgres

Postgres URLs are connection strings with the `postgres://` protocol. The most
notable URL query string parameter is `ssl=true`, which will turn on SSL.  An
example of a valid URL is `postgres://localhost:5432/database`.

Postgres sources support authentication types none and postgres.  If postgres
is used, specify `userPasswordCredentials`.

Postgres queries support parameterization like Influx queries, but the
parameters are numbered instead of named. `$1` is the beginning of the time interval
(inclusive) and `$2` the end (exclusive).

Queries should return time and value where time is a timestamp type and value
is a non-null numeric type.

Example:

```
select time, value from test where time >= $1 and time < $2
```

### Redshift

Read this section carefully.  There's a lot of details.

Redshift is Postgres 8 with a bunch of Amazon extensions and limitations for
big data.  Queries are consequently mostly the same as Postgres (i.e. use `$1`
and `$2`, it's SQL).  Whoever is configuring it should be aware of the
differences.  Configuring it is different, however, and you are encouraged to
read `test_scripts/provisioners/redshift.js` for sample working requests.

In order to uniquely identify and authenticate with a cluster, you need
the following:

- Source type `redshift`.
- Authentication type `aws`
- In `source.awsCredentials`:
  - `accessKeyId` and `secretAccessKey` from the standard AWS IAM credentials.
    This user must have permission to call
    [GetClusterCredentials])https://docs.aws.amazon.com/redshift/latest/APIReference/API_GetClusterCredentials.html).
    Whoever is responsible for handling AWS devops should know how to do this.
- In `source.redshiftConfiguration`:
  - `region`: The AWS region. We can't autodetect it.
  - `clusterId`:  The cluster identifier.
  - `dbUser`: the user we'll connect to the database as.  Must be
    preconfigured.  While AWS supports autocreating, we can't use it because it
    requires too much knowledge about their schema.

Clusters in Redshift have an endpoint, for example:
`metrics-importer-dev.cuhkbtf22f1f.us-east-2.redshift.amazonaws.com:5439/dev`.

This needs to be converted to a postgres url by prepending `postgres://` and
used as `source.url`.  This is because the AWS API is somewhat slow and likes
to have quotas.  By using this and an internal credentials cache, we can
avoid touching AWS's management APIs save once per source per 15 minutes.

## Manual Testing

Until we get a story around E2E testing more globally, you'll want to manually
test against the relevant backends. To aid in doing so, there are two sets of
test scripts:

- `test_scripts/data/*` knows how to upload data into a backend of a given type.
- `test_scripts/provisioners/*` knows how to provision a source and an
  importer.

At the moment provisioners assume that there is a metric_group "stability" and
metric "metric" on a coronerd with a project "cts".  They also don't share
sources.

Data scripts generate a month of random values every minute for the past month
from now.

Comments at the top of these scripts explain the usage.

All scripts need two tokens, an API token with capability `error:post` and a
user session token. The easiest way to get a user session token is
`morgue login`, then:

```
token=$( cat ~/.morgue/current.json | jq -r .config.token)
```

Which will give you  a `$token` to use in your bash commands.

Finally, interesting logs go to the sqlite3 database. Opening an sqlite3 shell
and doing `select message from logMessage order by time` with filters and
other qualifiers is a convenient way to tail the logs without a UI.

## Notes to devs

We use `prettier` for formatting. Before opening diffs, do:

```
npm run format
```

To add a new scraper type, edit `entities.ScraperType` to have it, then in
`scrapers/myscraper.ts` implement the interfaces from `scrapers/interface.ts`.
Then register your class in `src/scraperLoop.ts`.

Note that `src/authenticatedAxios.ts` offers an Axios wrapper that can inject
authentication into requests based off database configuration.

If you extend the schema, primary keys should always be `id` and a uuid,
unless the object is potentially very high volume.

It should be noted that the database uses MS since the Unix epoch for
timestamps.  Please follow this convention. Sqlite date handling is basically
broken even before bringing typeorm into the mix; typeorm + sqlite + dates
means a lot of subtle places for things to go wrong.  Using only seconds for
i.e. log messages means a (much more) ambiguous order, andMS since Unix epoch
matches `Date.now()` and `Date.getTime()` as well.

Most of the magic happens in `src/scraperLoop.ts`. We sync sources and importers
to it in a declaratifve fashion, and it pumps the queue. At the moment it's
single-process single-threaded, but if/when this needs to scale we'll have to
reimplement it.  Much mileage for scalability can be had via the `worker_thread`
module, but note that going multi-process or multi-machine will require
either a really good queueing solution or work in coronerd to protect against
duplicate submission.

### Scraper loop Scheduling Details

The scraper loop schedules up to a hardcoded number of scrapers at a time
(currently 20) with jitter, linear backoff, and an error limiter.

Coronerd requests are retried up to 5 times for 5xx status codes. All other
status codes are thrown and captured in the scraper's logs.

## Detecting Unhealthy States

The service publishes a number of scraper loop metrics.  All of the following
are unexpected states that may be hit momentarily but which should never be
hit for any length of time:

- `activeScrapers` is 0, `pendingScrapers` shows work.
- `scraperLoopIterations` stops incrementing.
- Metrics indicate work in the queues from more than a couple scrapers but
  `processedDataPoints` isn't increasing.

## Emergency SQL commands

Assumes sqlite on the metrics-importer database and that the service isn't
running:

### Disable all importers

```
UPDATE importer SET enabled = 0;
```

### Disable a bad importer

```
UPDATE importer SET enabled = 0 where id = 'badimporter-id';
```

### Stop a backfilling importer

Find the current unix time. Multiply by 1000. Then put it
in the following query as `now`:

```
UPDATE importer SET (earliestScraped, latestScraped) = (now, now)
where id = 'bad-importer-id';
```